/*
 * This file is part of the ESO Toolkit
 * Copyright (C) 2016 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef ESOTK_IDP_H_
#define ESOTK_IDP_H_

/*----------------------------------------------------------------------------*/
/**
 *                              Includes
 */
/*----------------------------------------------------------------------------*/

#include <cpl.h>

/*----------------------------------------------------------------------------*/
/**
 *                              Defines
 */
/*----------------------------------------------------------------------------*/

/* IDP KEYWORDS: PRIMARY HEADER */
#define ESOTK_IDP_PRI_SIMPLE               "SIMPLE"
#define ESOTK_IDP_PRI_BITPIX               "BITPIX"
#define ESOTK_IDP_PRI_NAXIS                "NAXIS"
#define ESOTK_IDP_PRI_EXTEND               "EXTEND"
#define ESOTK_IDP_PRI_ORIGIN               "ORIGIN"
#define ESOTK_IDP_PRI_TELESCOP             "TELESCOP"
#define ESOTK_IDP_PRI_INSTRUME             "INSTRUME"
#define ESOTK_IDP_PRI_DISPELEM             "DISPELEM"
#define ESOTK_IDP_PRI_SPECSYS              "SPECSYS"
#define ESOTK_IDP_PRI_OBJECT               "OBJECT"
#define ESOTK_IDP_PRI_EXT_OBJ              "EXT_OBJ"
#define ESOTK_IDP_PRI_RA                   "RA"
#define ESOTK_IDP_PRI_DEC                  "DEC"
#define ESOTK_IDP_PRI_EQUINOX              "EQUINOX"
#define ESOTK_IDP_PRI_RADESYS              "RADESYS"
#define ESOTK_IDP_PRI_EXPTIME              "EXPTIME"
#define ESOTK_IDP_PRI_TEXPTIME             "TEXPTIME"
#define ESOTK_IDP_PRI_MJD_OBS              "MJD-OBS"
#define ESOTK_IDP_PRI_MJD_END              "MJD-END"
#define ESOTK_IDP_PRI_PROG_ID              "PROG_ID"        /* Used as a regular expression */
#define ESOTK_IDP_PRI_PROG_ID_MULTI        "MULTI"          /* Not checked, it's used in the PROG_ID keyword to mark the origin in multiples (PROG_ID, PROGID) */
#define ESOTK_IDP_PRI_PROGID               "PROGID"         /* Not checked in the list of keywords, used in the PROG_ID keyword */
#define ESOTK_IDP_PRI_OBID                 "OBID"           /* Used as a regular expression */
#define ESOTK_IDP_PRI_M_EPOCH              "M_EPOCH"
#define ESOTK_IDP_PRI_PROV                 "PROV"           /* Used as a regular expression */
#define ESOTK_IDP_PRI_PROCSOFT             "PROCSOFT"
#define ESOTK_IDP_PRI_OBSTECH              "OBSTECH"
#define ESOTK_IDP_PRI_PRODCATG             "PRODCATG"
#define ESOTK_IDP_PRI_FLUXCAL              "FLUXCAL"
#define ESOTK_IDP_PRI_CONTNORM             "CONTNORM"
#define ESOTK_IDP_PRI_WAVELMIN             "WAVELMIN"
#define ESOTK_IDP_PRI_WAVELMAX             "WAVELMAX"
#define ESOTK_IDP_PRI_SPEC_BIN             "SPEC_BIN"
#define ESOTK_IDP_PRI_TOT_FLUX             "TOT_FLUX"
#define ESOTK_IDP_PRI_FLUXERR              "FLUXERR"
#define ESOTK_IDP_PRI_NCOMBINE             "NCOMBINE"       /* Not checked in the list of keywords, filled in keyword PROV */
#define ESOTK_IDP_PRI_REFERENC             "REFERENC"
#define ESOTK_IDP_PRI_SNR                  "SNR"
#define ESOTK_IDP_PRI_SNR_REDUCED          "SNR_REDUCED"    /* Not checked in the list of keywords, used in keyword SNR if this column not exist    */
#define ESOTK_IDP_PRI_BPM                  "BPM"            /* Not checked in the list of keywords, used in keyword SNR for search the BPM column   */
#define ESOTK_IDP_PRI_BPM_REDUCED          "BPM_REDUCED"    /* Not checked in the list of keywords, used in keyword SNR if the BPM column not exist */
#define ESOTK_IDP_PRI_SPEC_RES             "SPEC_RES"
#define ESOTK_IDP_PRI_PRO_CATG             "ESO PRO CATG"
#define ESOTK_IDP_PRI_CHECKSUM             "CHECKSUM"
#define ESOTK_IDP_PRI_DATASUM              "DATASUM"

/* IDP KEYWORDS: EXTENSION HEADER */
#define ESOTK_IDP_EXT_XTENSION             "XTENSION"
#define ESOTK_IDP_EXT_BITPIX               "BITPIX"
#define ESOTK_IDP_EXT_NAXIS                "NAXIS"
#define ESOTK_IDP_EXT_VOCLASS              "VOCLASS"
#define ESOTK_IDP_EXT_VOPUB                "VOPUB"
#define ESOTK_IDP_EXT_TITLE                "TITLE"
#define ESOTK_IDP_EXT_OBJECT               "OBJECT"
#define ESOTK_IDP_EXT_RA                   "RA"
#define ESOTK_IDP_EXT_DEC                  "DEC"
#define ESOTK_IDP_EXT_APERTURE             "APERTURE"
#define ESOTK_IDP_EXT_TELAPSE              "TELAPSE"
#define ESOTK_IDP_EXT_TMID                 "TMID"
#define ESOTK_IDP_EXT_SPEC_VAL             "SPEC_VAL"
#define ESOTK_IDP_EXT_SPEC_BW              "SPEC_BW"
#define ESOTK_IDP_EXT_TFIELDS              "TFIELDS"
#define ESOTK_IDP_EXT_NELEM                "NELEM"
#define ESOTK_IDP_EXT_TTYPE                "TTYPE*"         /* Regular expression */
#define ESOTK_IDP_EXT_TFORM                "TFORM*"         /* Regular expression, not used --> Put it by HDRL/ESOTK save function */
#define ESOTK_IDP_EXT_TUTYP                "TUTYP*"         /* Regular expression */
#define ESOTK_IDP_EXT_TUNIT                "TUNIT*"         /* Regular expression */
#define ESOTK_IDP_EXT_TUNIT1               "TUNIT1"         /* Not checked in the list of keywords, used to check the 'WAVE' units */
#define ESOTK_IDP_EXT_TUCD                 "TUCD*"          /* Regular expression */
#define ESOTK_IDP_EXT_TDMIN1               "TDMIN1"
#define ESOTK_IDP_EXT_TDMAX1               "TDMAX1"
#define ESOTK_IDP_EXT_EXTNAME              "EXTNAME"
#define ESOTK_IDP_EXT_INHERIT              "INHERIT"
#define ESOTK_IDP_EXT_CHECKSUM             "CHECKSUM"
#define ESOTK_IDP_EXT_DATASUM              "DATASUM"

/* Comments in the description columns keywords */
#define ESOTK_IDP_EXT_TTYPE_COMMENT        "Label for field for column-ith"
#define ESOTK_IDP_EXT_TFORM_COMMENT        "Data format of field for column-ith"
#define ESOTK_IDP_EXT_TUTYP_COMMENT        "IVOA data model element"
#define ESOTK_IDP_EXT_TUNIT_COMMENT        "Physical unit of field for column-ith"
#define ESOTK_IDP_EXT_TUCD_COMMENT         "UCD"

/* Valid 'WAVE' units */
#define ESOTK_SPECTRUM1D_TUNIT_NM          "nm"
#define ESOTK_SPECTRUM1D_TUNIT_ANGSTROM    "angstrom"

/*----------------------------------------------------------------------------*/
/**
 *                              Data structures
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 *                              Functions prototypes
 */
/*----------------------------------------------------------------------------*/

CPL_BEGIN_DECLS

/* Fill the primary and the extension header with keyword necessary to IDP standard */
cpl_error_code esotk_idp_fill_headers(
    cpl_frameset       *frameset,
    const cpl_table    *table,
    cpl_propertylist   *header_pri,
    cpl_propertylist   *header_ext,
    cpl_boolean        deleteColumnDefinition,
    cpl_size           nKeys,
    char               **enum_keys);

/* Update the keywords in the headers related with the 'WAVE' units */
cpl_error_code esotk_idp_update_wave_unit(
    cpl_propertylist   *header_pri,
    cpl_propertylist   *header_ext);

CPL_END_DECLS


#endif /* ESOTK_IDP_H_ */
