/*
 * This file is part of the ESO Toolkit
 * Copyright (C) 2016 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef ESOTK_UTILS_H_
#define ESOTK_UTILS_H_

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "hdrl.h"

#include <cpl.h>
#include <string.h>
#include <math.h>

CPL_BEGIN_DECLS

/*-----------------------------------------------------------------------------
                                Functions prototypes
 -----------------------------------------------------------------------------*/

cpl_error_code esotk_save_image( const char              *procatg,
								 const char              *recipe,
								 const char              *filename,
								 const cpl_type          save_type,
								 const cpl_image         *image,
								 const cpl_table         *qctable,
								 const cpl_parameterlist *parlist,
								 cpl_frameset            *frameset,
								 const char              *type,
								 const cpl_frame         *inherit);

int esotk_get_naxis1(const cpl_propertylist *plist);
int esotk_get_naxis2(const cpl_propertylist *plist);

/* Check inputs in the recipe */
cpl_error_code esotk_check_and_set_groups(
  cpl_frameset *frameset);

/* Clean parameters allocated in the recipe */
void esotk_clean_recipe_parameters(
	cpl_frameset *fs, cpl_frameset *fs_cnf, cpl_frameset *fs_bpm,
	cpl_frameset *fs_obj, cpl_frameset *fs_err, cpl_frameset *fs_mask,
	cpl_frame *f_data, cpl_frame *f_err, cpl_frame *f_bpm,
	cpl_mask* mask, hdrl_parameter *p);

/* Check all of numbers of things in the framesets */
cpl_error_code esotk_integrity_checks(
	cpl_frameset *fs, cpl_frameset *fs_cnf, cpl_frameset *fs_bpm,
	cpl_frameset *fs_obj, cpl_frameset *fs_err, cpl_frameset *fs_mask,
	cpl_frame *f_data, cpl_frame *f_err, cpl_frame *f_bpm,
	cpl_mask *mask, hdrl_parameter *p);

/* Check if are the same cpl_frameset */
cpl_error_code esotk_checkmatch(const cpl_frameset *ref, const cpl_frameset *test);


CPL_END_DECLS

#endif /* ESOTK_UTILS_H_ */
