/*
 * This file is part of the HDRLDEMO pipeline
 * Copyright (C) 2013 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _DEFAULT_SOURCE
#define _BSD_SOURCE

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 *                              Defines
 */
/*----------------------------------------------------------------------------*/

#define RECIPE_NAME                                         "esotk_spectrum1d_combine"

#define ESOTK_SPECTRUM1D_SNR_TUNIT_VALUE                    ""
#define ESOTK_SPECTRUM1D_SNR_TUTYP_VALUE                    "eso:Data.FluxAxis.Accuracy.SNR"
#define ESOTK_SPECTRUM1D_SNR_TUCD_VALUE                     "stat.snr;meta.main"

#define ESOTK_SPECTRUM1D_SNR_REDUCED_TUNIT_VALUE            ""
#define ESOTK_SPECTRUM1D_SNR_REDUCED_TUTYP_VALUE            "eso:Data.FluxAxis.Accuracy.SNR"
#define ESOTK_SPECTRUM1D_SNR_REDUCED_TUCD_VALUE             "stat.snr"

#define ESOTK_SPECTRUM1D_BPM_TTYPE_VALUE                    "QUAL"
#define ESOTK_SPECTRUM1D_BPM_TUNIT_VALUE                    ""
#define ESOTK_SPECTRUM1D_BPM_TUTYP_VALUE_VOCLASS_v1         "Spectrum.Data.FluxAxis.Quality"
#define ESOTK_SPECTRUM1D_BPM_TUTYP_VALUE_VOCLASS_v2         "spec:Data.FluxAxis.Accurancy.QualityStatus"
#define ESOTK_SPECTRUM1D_BPM_TUCD_VALUE                     "meta.code.qual;meta.main"

#define ESOTK_SPECTRUM1D_BPM_REDUCED_TTYPE_VALUE            "QUAL_REDUCED"
#define ESOTK_SPECTRUM1D_BPM_REDUCED_TUNIT_VALUE            ""
#define ESOTK_SPECTRUM1D_BPM_REDUCED_TUTYP_VALUE_VOCLASS_v1 "Spectrum.Data.FluxAxis.Quality"
#define ESOTK_SPECTRUM1D_BPM_REDUCED_TUTYP_VALUE_VOCLASS_v2 "spec:Data.FluxAxis.Accurancy.QualityStatus"
#define ESOTK_SPECTRUM1D_BPM_REDUCED_TUCD_VALUE             "meta.code.qual"


#define ESOTK_SPECTRUM1D_CONTRIB_TUNIT_VALUE                ""
#define ESOTK_SPECTRUM1D_CONTRIB_TUTYP_VALUE                ""
#define ESOTK_SPECTRUM1D_CONTRIB_TUCD_VALUE                 "meta.number"

#define HDRL_USE_EXPERIMENTAL

/*----------------------------------------------------------------------------*/
/**
 *                              Includes
 */
/*----------------------------------------------------------------------------*/

#include <string.h>
#include <ctype.h>


#include <cpl.h>

#include "hdrl.h"

#include "esotk_dfs.h"
#include "esotk_utils.h"
#include "esotk_idp.h"

/*----------------------------------------------------------------------------*/
/**
 *                 Typedefs: Enumeration types
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 *                 Defines
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 *                 Macros
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 *                 Typedefs: Structured types
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 *                 Functions prototypes
 */
/*----------------------------------------------------------------------------*/

/* Function to save the table result in a FITS file */
static cpl_error_code save_tab(
    cpl_table                 *spectrum_table,
    const cpl_parameterlist   *parlist,
    cpl_frameset              *frameset,
    cpl_propertylist          *qclist,
    cpl_frame                 *raw_frame,
    cpl_boolean               isIDP,
    const char                *filename,
    const char                *catg,
    cpl_size                  nKeys,
    char                      **copy_keys,
    cpl_size                  nOutputs,
    char                      **output_columns);

/* Convert array to image */
static cpl_image * arr_to_img(
    const cpl_array           *arr,
    const cpl_array           *rej);

/* Convert input arrays in a spectrum */
static hdrl_spectrum1D * convert_arrs_to_spectrum1D(
    const cpl_array           *wlens,
    const cpl_array           *flx_arr,
    const cpl_array           *flx_e_arr,
    const cpl_array           *quality);

/* Convert table to spectrum */
static hdrl_spectrum1D * convert_to_spectrum(
    const cpl_table           *tb,
    const char                *wln_name,
    const char                *flx_name,
    const char                *flx_e_name,
    const char                *qual_column);

/* Generate a hdrl_spectrum1Dlist from the input frameset */
static hdrl_spectrum1Dlist * generate_list_from_frameset(
    const cpl_frameset        *fset,
    const cpl_boolean         noIDP,
    const int                 extension,
    const char                *wavelength_column,
    const char                *flux_column,
    const char                *flux_error_column,
    const char                *qual_column,
    const cpl_boolean         rescale,
    cpl_vector                *rescale_vector);

/* Generate wavelenght array form the input frameset */
static cpl_array * generate_wlenghts_array_from_frameset(
    const cpl_frameset        *fset);

/* Get wavelenghts from a destination frameset */
static cpl_array * get_wlengths(
    const cpl_frameset        *fset,
    const hdrl_spectrum1Dlist *list);

/* Collapse spectrum 1D list */
static cpl_error_code hdrl_spectrum1Dlist_collapse_esotk(
    const hdrl_spectrum1Dlist *list,
    const hdrl_parameter      *stacking_par,
    const cpl_array           *wlengths,
    const hdrl_parameter      *resample_par,
    const hdrl_parameter      *bpm_par,
    const cpl_boolean         mark_bpm_in_interpolation,
    hdrl_spectrum1D           **result,
    cpl_image                 **contrib,
    hdrl_imagelist            **resampled_and_aligned_fluxes);

/* Function that collapse all of spectra in the input list */
static void collapse_spectra(
    const hdrl_spectrum1Dlist *list,
    const cpl_array           *wlengths,
    const cpl_boolean         remove_rej,
    const hdrl_parameter      *collapse_par,
    const hdrl_parameter      *interp_par,
    const hdrl_parameter      *bpm_par,
    hdrl_spectrum1D           **arg_spectrum1D,
    cpl_image                 **arg_contrib);

/* Function that collapse all of spectra in the input list and convert in a table */
static cpl_table * collapse_and_convert_to_table(
    const hdrl_spectrum1Dlist *list,
    const cpl_array           *wlengths,
    const cpl_boolean         remove_rej,
    const hdrl_parameter      *collapse_par,
    const hdrl_parameter      *interp_par,
    const hdrl_parameter      *bpm_par,
    const char                *wav_cl_name,
    const char                *flx_cl_name,
    const char                *flx_e_cl_name,
    const char                *bpm_cl_name,
    const char                *contrib_cl_name,
    const char                *snr_cl_name);

/* Add columns to the input table */
static void append_table(
    cpl_table                 *t,
    const cpl_table           *other);

/* Add columns that match in the array string list and add the correspondence keywords to the header */
static cpl_table * append_columns_table(
    cpl_frameset              *frameset,
    const cpl_table           *source,
    cpl_size                  size,
    char                      **out_cols,
    cpl_propertylist          *pri_header,
    cpl_propertylist          *ext_header);


/* Combine two input tables */
static cpl_table * attach_tables(
    const cpl_table           *t1,
    const cpl_table           *t2);

/* Convert a column in the input table to array */
static cpl_array * convert_to_array(
    const cpl_table           *tb,
    const char                *cl_name);

/* Change table NxM by Nx2, where the first row are names and the second row are data arrays */
static cpl_table * convert_to_idp(
    const cpl_table           *tb);

/* Split the input str string using the delimiter delim. Also get the number of tokens */
static char ** strsplit(
    cpl_boolean               isKeyword,
    const char                *str,
    const char                *delim,
    cpl_size                  *numtokens);

/* Trim the input str with spaces and replace all characters c_orig by c_replace */
static char * str_trim_and_replace(
    char                      *str,
    const char                c_orig,
    const char                c_replace);

cpl_error_code
esotk_spectrum1d_qc_fromheader(cpl_frameset * frameset, const char * key_name,
			       cpl_propertylist * qclist);

cpl_error_code
esotk_spectrum1d_compute_qc(cpl_table *tb_output_idp,
			    cpl_frameset *frameset_flux,
			    cpl_vector *rescale_vector,
			    cpl_propertylist *qclist);

/*----------------------------------------------------------------------------*/
/**
 *                          Static variables
 */
/*----------------------------------------------------------------------------*/

static hdrl_spectrum1D_wave_scale global_scale = hdrl_spectrum1D_wave_scale_linear;

static char esotk_spectrum1d_combine_description[] =
"This recipe combines a list of input 1D spectra into a single spectrum.  The    \n"
"user can provide the combining method and the interpolation method.  If no      \n"
"destination wavelengths are specified, the wavelengths of the first spectrum are\n"
"used instead. SPECTRUM_1D must be a table.                                      \n"
"                                                                                \n"
"If IDP compliance is strictly enforced (--noIDP=FALSE), the FLUX(_REDUCED)      \n"
"column must contain the spectrum flux, the ERR(_REDUCED) column the spectrum    \n"
"error and the WAVE column the wavelengths. The QUAL column (optional and        \n"
"controlled by --use-quality-column) indicates a bad pixel if the value is       \n"
"different from 0.                                                               \n"
"                                                                                \n"
"If IDP compliance is not enforced (--noIDP=TRUE), the user has to provide the   \n"
"extension containing the spectrum table, and the names of the columns containing\n"
"the wavelength, the flux, the error of the flux, and the quality. Moreover, not \n"
"only the usage of the quality flag is optional (again controlled by             \n"
"--use-quality-column) but also the presence of a column containing the error    \n"
"(controlled by the presence or absence of the column name specified by          \n"
"--noIDP.colname-err).                                                           \n"
"                                                                                \n"
"Input files:                                                                    \n"
"                                                                                \n"
"   DO category:                 Explanation:              Required:             \n"
"   --------------------------   --------                  ---------             \n"
"   SPECTRUM_1D                  1D spectrum               Yes                   \n"
"   SPECTRUM_COLLAPSE_WLENGTHS   Destination wavelengths   No                    \n"
"                                                                                \n"
"Output files:                                                                   \n"
"                                                                                \n"
"   DO category:                 Explanation:                                    \n"
"   --------------------------   --------                                        \n"
"   ESOTK_SPECTRUM_COMBINED      The combined spectrum                           \n"
"   ESOTK_SPECTRUM_IDP_FORMAT    The combined spectrum in IDP format             \n";

/* Standard CPL recipe definition */
cpl_recipe_define( esotk_spectrum1d_combine,
                   ESOTK_BINARY_VERSION,
                   "HDRL Group",
                   PACKAGE_BUGREPORT,
                   "2017",
                   "Combines 1D sectra",
                   esotk_spectrum1d_combine_description);

/*----------------------------------------------------------------------------*/
/**
 *                 Functions
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup esotk_spectrum1d_combine Combining function for a list of 1D spectra
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @brief   perform spectrum 1D combine
 *
 * @param   frameset   input set of frames
 * @param   parlist    input recipe parameters
 *
 * @return  cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static int esotk_spectrum1d_combine(
    cpl_frameset            *frameset,
    const cpl_parameterlist *parlist)
{
  if (esotk_check_and_set_groups(frameset) != CPL_ERROR_NONE) {
      return cpl_error_get_code();
  }

  /* Temporary framesets */
  cpl_size     fsize                          = cpl_frameset_get_size(frameset);
  cpl_frameset *frameset_flux                 = cpl_frameset_new();
  cpl_frameset *frameset_destination_wlenghts = cpl_frameset_new();
  cpl_frame    *spectrum_frame                = NULL;


  /* Split the input frameset in diferent cpl_frameset's: "Raw Spectrum" or "Destination Wavelengths collapse" */
  for (cpl_size i = 0; i < fsize; i++) {

      cpl_frame *cur_frame = cpl_frameset_get_position(frameset, i);

      if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_SPECTRUM_1D)) {

          spectrum_frame = cpl_frame_duplicate(cur_frame);
          cpl_frameset_insert(frameset_flux, spectrum_frame);

      } else if (!strcmp(ESOTK_WAVELENGTHS_DEST_COLLAPSE, cpl_frame_get_tag(cur_frame))) {

          cpl_frameset_insert(frameset_destination_wlenghts, cpl_frame_duplicate(cur_frame));
          cpl_msg_info(cpl_func,"Destination wavelengths found in the SOF");

      } else {

          cpl_msg_info(cpl_func, "SOF contains unknown tags");
      }
  }

  /* Check if exist one or more raw spectrum */
  if (cpl_frameset_get_size(frameset_flux) < 1) {
      cpl_frameset_delete(frameset_flux);
      cpl_frameset_delete(frameset_destination_wlenghts);
      return cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                                   "SOF contains wrong number of frames");
  }
  /* Warn if there is only one raw spectrum */
  if (cpl_frameset_get_size(frameset_flux) == 1) {
      cpl_msg_warning(cpl_func,"Only one frame in the SOF !" );
  }


  /*** Get input parameter from the cpl_paramterlist ***/

  const cpl_parameter *par             = cpl_parameterlist_find_const(parlist, RECIPE_NAME".collapse.reject-bpm");
  const cpl_boolean   remove_rej       = cpl_parameter_get_bool(par);

  par             = cpl_parameterlist_find_const(parlist, RECIPE_NAME".bpm.enable");
  const cpl_boolean   identify_cosmics = cpl_parameter_get_bool(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".use-quality-column");
  const cpl_boolean   use_qty          = cpl_parameter_get_bool(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".rescale-spectra");
  const cpl_boolean   rescale          = cpl_parameter_get_bool(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".output-columns");
  const char*         str_output_colum = cpl_parameter_get_string(par);

  /* Check .output-columns parameter */
  cpl_size nOutputs = 0;
  char **output_columns = strsplit(CPL_FALSE, str_output_colum, ",", &nOutputs);
  for (int i = 0; i < nOutputs; i++) {
      if (!(   !strcmp(output_columns[i], "ALL"            )
            || !strcmp(output_columns[i], "QUAL"           )
            || !strcmp(output_columns[i], "QUAL_REDUCED"   )
            || !strcmp(output_columns[i], "BPM"            )
            || !strcmp(output_columns[i], "BPM_REDUCED"    )
            || !strcmp(output_columns[i], "CONTRIB"        )
            || !strcmp(output_columns[i], "CONTRIB_REDUCED")
            || !strcmp(output_columns[i], "ERR"            )
            || !strcmp(output_columns[i], "ERR_REDUCED"    )
            || !strcmp(output_columns[i], "FLUX"           )
            || !strcmp(output_columns[i], "FLUX_REDUCED"   )
            || !strcmp(output_columns[i], "SNR"            )
            || !strcmp(output_columns[i], "SNR_REDUCED"    )
            || !strcmp(output_columns[i], "WAVE"           ) )
          || ( !strcmp(output_columns[i], "ALL") && i > 0  ) ){

          cpl_frameset_delete(frameset_flux);
          cpl_frameset_delete(frameset_destination_wlenghts);
          if (!strcmp(output_columns[i], "ALL")) {
              for (cpl_size token = 0; token < nOutputs; token++) {
                  if (output_columns[token]) cpl_free(output_columns[token]);
              }
              cpl_free(output_columns);
              return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, ".output-columns = 'ALL' non-valid, need to be unique!");
          } else {
              for (cpl_size token = 0; token < nOutputs; token++) {
                  if (output_columns[token]) cpl_free(output_columns[token]);
              }
              cpl_free(output_columns);
              return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, ".output-columns non-valid!");
          }
      }
  }

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".copykeys");
  const char *str_copy_keys = cpl_parameter_get_string(par);
  cpl_size nKeys = 0;
  char **copy_keys = strsplit(CPL_TRUE, str_copy_keys, ",", &nKeys);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".noIDP");
  const cpl_boolean noIDP = cpl_parameter_get_bool(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".noIDP.extension");
  const int   extension = cpl_parameter_get_int(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".noIDP.colname-wave");
  const char*         colname_wave = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".noIDP.colname-flux");
  const char*        colname_flux  = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".noIDP.colname-err");
  const char*     colname_err     = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".noIDP.colname-qual");
  const char*        colname_qual  = cpl_parameter_get_string(par);



  const char          *qty_cl_name  = NULL;
  hdrl_spectrum1Dlist *list         = NULL;
  hdrl_spectrum1Dlist *list_reduced = NULL;
  cpl_vector * rescale_vector = NULL;
  cpl_vector * rescale_vector_reduced = NULL;

  rescale_vector = cpl_vector_new(cpl_frameset_get_size(frameset_flux));
  rescale_vector_reduced = cpl_vector_new(cpl_frameset_get_size(frameset_flux));
  cpl_vector_fill(rescale_vector, 0.);
  cpl_vector_fill(rescale_vector_reduced, 0.);

  if (noIDP == CPL_FALSE) {
      /* Generate list of spectrums, depending of if the user want use the quality column */
      qty_cl_name  = use_qty ? "QUAL" : NULL;
      list         = generate_list_from_frameset(frameset_flux, noIDP, 1, "WAVE", "FLUX",         "ERR",         qty_cl_name, rescale, rescale_vector);
      list_reduced = generate_list_from_frameset(frameset_flux, noIDP, 1, "WAVE", "FLUX_REDUCED", "ERR_REDUCED", qty_cl_name, rescale, rescale_vector_reduced);
  } else {
      /* Generate list of spectrums, depending of if the user want use the quality column */
      qty_cl_name  = use_qty ? colname_qual : NULL;
      list         = generate_list_from_frameset(frameset_flux, noIDP, extension, colname_wave, colname_flux, colname_err, qty_cl_name, rescale, rescale_vector);
      list_reduced = NULL; /* We only use the one provided by the user */
  }
  /* Check the size of the frame list spectrums */
  const hdrl_spectrum1Dlist *not_null_list = (list != NULL) ? list : list_reduced;
  if(!not_null_list || cpl_error_get_code()){

      cpl_frameset_delete(frameset_flux);
      cpl_frameset_delete(frameset_destination_wlenghts);
      cpl_vector_delete(rescale_vector);
      cpl_vector_delete(rescale_vector_reduced);

      for (cpl_size token = 0; token < nOutputs; token++) {
          if (output_columns[token]) cpl_free(output_columns[token]);
      }
      cpl_free(output_columns);

      if (copy_keys) {
        for (cpl_size i = 0; i < nKeys; i++) {
            if(copy_keys[i]) cpl_free(copy_keys[i]);
        }
        cpl_free(copy_keys);
      }

      return cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
				   "SOF contains wrong frames (number or names). "
				   "Must contain, at least 3 if IDP checks"
				   " are enabled ('WAVE','FLUX','ERR' or 'WAVE_REDUCED','FLUX_REDUCED','ERR_REDUCED'), "
				   "or 2 (wavelength and flux column), "
				   "if IDP checks are disabled");
  }

  /* Get all of the wavelengths in the not_null_list */
  cpl_array *wlengths = get_wlengths(frameset_destination_wlenghts, not_null_list);

  /* Generate hdrl_parameter to: Collapse and interpolation, based in the input recipe parameters */
  hdrl_parameter *collapse_par = hdrl_collapse_parameter_parse_parlist(                       parlist, RECIPE_NAME".collapse"     );
  hdrl_parameter *interp_par   = hdrl_spectrum1D_resample_interpolate_parameter_parse_parlist(parlist, RECIPE_NAME".interpolation");
  /* Parse the BPM Parameters */
  hdrl_parameter * bpm_params = NULL;

  if(identify_cosmics)
    bpm_params = hdrl_bpm_3d_parameter_parse_parlist(parlist, RECIPE_NAME".bpm") ;


  /* Generate the normal table */
  cpl_table *tb = collapse_and_convert_to_table(
      list,         wlengths, remove_rej, collapse_par, interp_par, bpm_params,
      "WAVE", "FLUX",          "ERR",         "BPM",         "CONTRIB",         "SNR"        );

  /* Generate the reduced table */
  cpl_table *tb_reduced = collapse_and_convert_to_table(
      list_reduced, wlengths, remove_rej, collapse_par, interp_par, bpm_params,
      "WAVE", "FLUX_REDUCED",  "ERR_REDUCED", "BPM_REDUCED", "CONTRIB_REDUCED", "SNR_REDUCED");

  /* Combine tables and generate the collapse IDP table */
  cpl_table *tb_output     = attach_tables(tb, tb_reduced);
  cpl_table *tb_output_idp = convert_to_idp(tb_output);

  /* Add QC parameters derived from the input frame headers */
  cpl_propertylist * qclist = cpl_propertylist_new();
  if (list != NULL) {
      esotk_spectrum1d_compute_qc (tb_output_idp, frameset_flux, rescale_vector, qclist);
  } else {
      esotk_spectrum1d_compute_qc (tb_output_idp, frameset_flux, rescale_vector_reduced,  qclist);
  }
  cpl_vector_delete(rescale_vector);
  cpl_vector_delete(rescale_vector_reduced);

  /* Save normal output table */
  save_tab(tb_output,     parlist, frameset, qclist, spectrum_frame,
	   CPL_FALSE, "combined_spectrum.fits", "ESOTK_SPECTRUM_COMBINED", 0, NULL, 0, NULL);

  if (noIDP == CPL_FALSE) { /* Save only in IDP case */
      /* Save IDP    output table */
      save_tab(tb_output_idp, parlist, frameset, qclist, spectrum_frame,
	       CPL_TRUE, "combined_spectrum_idp_format.fits", "ESOTK_SPECTRUM_IDP_FORMAT",
	       nKeys, copy_keys, nOutputs, output_columns);
  }

  /*** Cleanup ***/
  cpl_propertylist_delete(qclist);
  cpl_frameset_delete(frameset_flux);
  cpl_frameset_delete(frameset_destination_wlenghts);
  hdrl_spectrum1Dlist_delete(list);
  hdrl_spectrum1Dlist_delete(list_reduced);
  cpl_table_delete(tb);
  cpl_table_delete(tb_reduced);
  cpl_table_delete(tb_output);
  cpl_table_delete(tb_output_idp);
  hdrl_parameter_delete(collapse_par);
  hdrl_parameter_delete(interp_par);
  hdrl_parameter_delete(bpm_params);
  cpl_array_delete(wlengths);
  for (cpl_size token = 0; token < nOutputs; token++) {
      if (output_columns[token]) cpl_free(output_columns[token]);

  }
  if (output_columns) cpl_free(output_columns);
  if (copy_keys) {
    for (cpl_size i = 0; i < nKeys; i++) {
        if(copy_keys[i]) cpl_free(copy_keys[i]);
    }
    cpl_free(copy_keys);
  }

  return (int)cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Compute QC parameters
 *
 * @param   tb_output_idp    The input table in idp format
 * @param   frameset_flux    The input frameset with the spectra to combine
 * @param   rescale_vector   The vector containing the absolute scaling factors
 * @param   qclist           The output propertylist with the QC parameter
 *
 * @return  cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/

cpl_error_code
esotk_spectrum1d_compute_qc(cpl_table *tb_output_idp,
			    cpl_frameset *frameset_flux,
			    cpl_vector *rescale_vector,
			    cpl_propertylist *qclist)
{

  cpl_ensure_code(tb_output_idp, CPL_ERROR_NULL_INPUT);
  cpl_ensure_code(frameset_flux, CPL_ERROR_NULL_INPUT);
  cpl_ensure_code(rescale_vector, CPL_ERROR_NULL_INPUT);
  cpl_ensure_code(qclist, CPL_ERROR_NULL_INPUT);

  cpl_msg_info(cpl_func, "Deriving QC parameter ...");
  cpl_msg_indent_more();

  cpl_msg_info(cpl_func, "Averaging input QC header parameter over the frames");
  esotk_spectrum1d_qc_fromheader (frameset_flux, "ESO QC AIRM AVG", qclist);
  esotk_spectrum1d_qc_fromheader (frameset_flux, "ESO QC IA FWHM AVG", qclist);
  esotk_spectrum1d_qc_fromheader (frameset_flux, "ESO QC IWV AVG", qclist);
  esotk_spectrum1d_qc_fromheader (frameset_flux, "SNR", qclist);
  /* Get SNR keyword from the first frame */
  double snr_inputframes_mean = -1;
  if (cpl_propertylist_has (qclist, "SNR"))
    {
      snr_inputframes_mean = cpl_propertylist_get_double (qclist, "SNR");
      cpl_propertylist_erase(qclist, "SNR");
    }

  cpl_msg_info(cpl_func, "Deriving CONTRIB and SNR related QC parameter from the"
      " recipe product:");

  /* Add QC parameters derived from idp table product */
  cpl_array *contrib_array = NULL;
  cpl_array *snr_array = NULL;
  const cpl_array *bpm_array = NULL;
  if (cpl_table_has_column (tb_output_idp, "CONTRIB")
      && cpl_table_has_column (tb_output_idp, "SNR")
      && cpl_table_has_column (tb_output_idp, "BPM"))
    {
      cpl_msg_info(cpl_func, "Deriving QC parameter from normal IDP columns, "
	  "not from the _REDUCED columns");
      contrib_array = cpl_array_duplicate(cpl_table_get_array (tb_output_idp, "CONTRIB", 0));
      snr_array = cpl_array_duplicate(cpl_table_get_array (tb_output_idp, "SNR", 0));
      /* QUAL column at this stage is still called BPM */
      bpm_array = cpl_table_get_array (tb_output_idp, "BPM", 0);
    }
  else if (cpl_table_has_column (tb_output_idp, "CONTRIB_REDUCED")
      && cpl_table_has_column (tb_output_idp, "SNR_REDUCED")
      && cpl_table_has_column (tb_output_idp, "BPM_REDUCED"))
    {
      cpl_msg_info(cpl_func, "Deriving QC parameter from _REDUCED columns");
      contrib_array = cpl_array_duplicate(cpl_table_get_array (tb_output_idp, "CONTRIB_REDUCED", 0));
      snr_array = cpl_array_duplicate(cpl_table_get_array (tb_output_idp, "SNR_REDUCED", 0));
      /* QUAL column at this stage is still called BPM */
      bpm_array = cpl_table_get_array (tb_output_idp, "BPM_REDUCED", 0);
    }

  if (contrib_array != NULL && snr_array != NULL && bpm_array != NULL)
    {
      /*Take bad pixels into account*/
      for (cpl_size i = 0; i < cpl_array_get_size (contrib_array); i++)
	{
	  if (cpl_array_get (bpm_array, i, NULL) > 0)
	    {
	      cpl_array_set_invalid (contrib_array, i);
	      cpl_array_set_invalid (snr_array, i);
	    }
	}
      cpl_msg_indent_more();
      cpl_msg_info(cpl_func, "Excluding Bad Pixels for all QC parameter");
      double contrib_mean = cpl_array_get_mean(contrib_array);
      double contrib_mean_rel = contrib_mean / cpl_frameset_get_size (frameset_flux);
      cpl_propertylist_append_double (qclist, "ESO QC CONTRIB AVG", contrib_mean);
      cpl_propertylist_set_comment (qclist, "ESO QC CONTRIB AVG", "mean contribution");
      cpl_propertylist_append_double (qclist, "ESO QC CONTRIB REL", contrib_mean_rel);
      cpl_propertylist_set_comment (qclist, "ESO QC CONTRIB REL", "relative mean contribution");

      /* Take pixels with SNR <= 0 NOT into account -> requested by QC group */
      for (cpl_size i = 0; i < cpl_array_get_size (snr_array); i++)
	{
	  if (cpl_array_get (snr_array, i, NULL) <= 0.)
	    {
	      cpl_array_set_invalid (snr_array, i);
	    }
	}
      cpl_msg_info(cpl_func, "Additionally excluding SNR <= 0 values for SNR "
	  "related QC parameter");
      double snr_mean = -1.;
      double snr_ratio = -1.;
      double snr_improv = -1.;

      if (cpl_array_count_invalid(snr_array) < cpl_array_get_size(snr_array)) {
	  snr_mean = cpl_array_get_mean(snr_array);
	  snr_ratio = snr_mean / snr_inputframes_mean;
	  snr_improv = snr_ratio / sqrt (cpl_frameset_get_size (frameset_flux));
      }

      cpl_propertylist_append_double (qclist, "ESO QC SNR AVG", snr_mean);
      cpl_propertylist_set_comment (qclist, "ESO QC SNR AVG", "mean Signal-to-Noise Ratio");
      cpl_propertylist_append_double (qclist, "ESO QC SNR RATIO", snr_ratio);
      cpl_propertylist_set_comment (qclist, "ESO QC SNR RATIO", "RATIO(SNR-prod, SNR-input-mean)");
      cpl_propertylist_append_double (qclist, "ESO QC SNR IMPROV", snr_improv);
      cpl_propertylist_set_comment (qclist, "ESO QC SNR IMPROV", "RATIO(SNR-ratio, sqrt(N))");
      cpl_propertylist_append_double (qclist, "ESO QC SQRTNCOM", sqrt(cpl_frameset_get_size (frameset_flux)));
      cpl_propertylist_set_comment (qclist, "ESO QC SQRTNCOM", "sqrt(N)");
      cpl_array_delete (contrib_array);
      cpl_array_delete (snr_array);
    }

    double scaling_mean = cpl_vector_get_mean(rescale_vector);
    double max_dev = 0.0;
    for (cpl_size i = 0; i < cpl_vector_get_size(rescale_vector); i++){
	double deviation =
	    fabs((cpl_vector_get(rescale_vector, i) - scaling_mean)/fabs(scaling_mean));
	if (deviation > max_dev) max_dev = deviation;
    }
    cpl_propertylist_append_double (qclist, "ESO QC MAXDEV", max_dev);
    cpl_propertylist_set_comment (qclist, "ESO QC MAXDEV", "relative maximum scaling deviation");

  cpl_msg_indent_less();
  return cpl_error_get_code();
}



/*----------------------------------------------------------------------------*/
/**
 * @brief   Compute some QC parameters from the header of the input frames
 *
 * @param   frameset    The input frameset
 * @param   key_name    The header keyword to average
 * @param   qclist      The output propertylist with the QC parameter
 *
 * @return  cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/

cpl_error_code
esotk_spectrum1d_qc_fromheader(cpl_frameset * frameset, const char * key_name,
		 cpl_propertylist * qclist)
{
  cpl_size  nframes = cpl_frameset_get_size(frameset);

  /* cpl_array_new(): All array elements are initially flagged as invalid. */
  cpl_array * qcarray = cpl_array_new(nframes, CPL_TYPE_DOUBLE);
  char * comment = NULL;
  int comment_set = 0;

  for (cpl_size i = 0; i < nframes ; i++) {
      cpl_frame *cur_frame = cpl_frameset_get_position(frameset, i);
      cpl_propertylist * plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
      if(cpl_propertylist_has(plist, key_name)){
	  cpl_array_set(qcarray, i , cpl_propertylist_get_double(plist, key_name));
      }
      if(cpl_propertylist_has(plist, key_name) && comment_set == 0){ /* Only once needed*/
	  comment = cpl_sprintf("%s", cpl_propertylist_get_comment(plist, key_name));
	  comment_set = 1;
      }

      cpl_propertylist_delete(plist);
  }

  if (cpl_array_count_invalid(qcarray) < nframes ){ /* Write QC */
      cpl_propertylist_append_double(qclist, key_name, cpl_array_get_mean(qcarray));
      cpl_propertylist_set_comment(qclist,key_name, comment);
  } else {
      cpl_propertylist_append_double(qclist, key_name, -1.0);
  }

  cpl_array_delete(qcarray);
  cpl_free(comment);
  return cpl_error_get_code();
}

/** @cond PRIVATE */

/*----------------------------------------------------------------------------*/
/**
 * @brief   Function needed by cpl_recipe_define to fill the input parameters
 *
 * @param   self         Output parameterlist
 *
 * @return  cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_spectrum1d_combine_fill_parameterlist(
    cpl_parameterlist *self)
{

  cpl_parameter *par;


  /* --use-quality-column */
  par = cpl_parameter_new_value(RECIPE_NAME".use-quality-column",	CPL_TYPE_BOOL,
                                "If TRUE and quality (QUAL column) exists and is not 0 the pixel is rejected. \n "
                                "If FALSE (or if the column does not exist) all the pixels are assumed as valid.",
                                RECIPE_NAME, CPL_FALSE);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "use-quality-column");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);


  /* --rescale-spectra */
  par = cpl_parameter_new_value(RECIPE_NAME".rescale-spectra", CPL_TYPE_BOOL,
                                "If TRUE all the spectra are multiplicatively rescaled \n "
                                "to the level of the first spectra in the SOF",
                                RECIPE_NAME, CPL_FALSE);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "rescale-spectra");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* --output-columns */
  par = cpl_parameter_new_value(RECIPE_NAME".output-columns",	CPL_TYPE_STRING,
                                " Comma separated string specifying the output columns.\n"
                                " The \"WAVE\" column can be omitted as it is a mandatory column. \n"
                                " The other columns can be selected (e.g. \"FLUX,ERR,SNR\").\n"
                                " If ALL is set, all columns computed by the recipe will be saved",
                                RECIPE_NAME, "ALL");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "output-columns");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* --copykeys */
  par = cpl_parameter_new_value(RECIPE_NAME".copykeys", CPL_TYPE_STRING,
                                " Comma separated string specifying a list of keywords.\n"
                                " If a keyword in this list is found in any of the input spectra, \n"
                                " the keyword will be propagated to the output spectrum \n"
                                " with the prefix HIERARCH ESO STACK and the number of the input file as suffix \n"
                                " (e.g. copykeys='ESO.SEQ.CUMOFFSETA,ESO.TEL.AIRM.START,ESO.TEL.AIRM.END,OBSERVER' --> \n"
                                "                 HIERARCH ESO STACK SEQ CUMOFFSETA1, HIERARCH ESO STACK SEQ CUMOFFSETA2, ...   \n"
                                "                 HIERARCH ESO STACK TEL AIRM START1, HIERARCH ESO STACK TEL AIRM START2, ...   \n"
                                "                 HIERARCH ESO STACK TEL AIRM END1,   HIERARCH ESO STACK TEL AIRM END2,   ...   \n"
                                "                 HIERARCH ESO STACK OBSERVER1,       HIERARCH ESO STACK OBSERVER2,       ... ) \n"
                                " If the keyword to copy is a HIERARCH ESO keyword, the HIERARCH ESO is stripped before appending\n",
                                RECIPE_NAME, " ");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "copykeys");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* Create interpolation parameters and add to the cpl_parameterlist */
  cpl_parameterlist *pintep = hdrl_spectrum1D_resample_interpolate_parameter_create_parlist(RECIPE_NAME, "interpolation", "AKIMA");
  for (cpl_parameter *p = cpl_parameterlist_get_first(pintep); p != NULL; p = cpl_parameterlist_get_next(pintep)) {
      cpl_parameterlist_append(self, cpl_parameter_duplicate(p));
  }
  cpl_parameterlist_delete(pintep);


  /* --collapse.reject-bpm */
  par = cpl_parameter_new_value(RECIPE_NAME".collapse.reject-bpm", CPL_TYPE_BOOL,
                                "If TRUE rejects an interpolated wavelength whose neighbor (in the original spectrum) is a bad pixel.\n "
                                "For the rejection to occur only one neighbor is sufficient.",
                                RECIPE_NAME, CPL_TRUE);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "collapse.reject-bpm");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);


  /* Create collapse parameters  and add to the cpl_parameterlist */
  hdrl_parameter    *sigclip_def = hdrl_collapse_sigclip_parameter_create(3., 3., 5);
  hdrl_parameter    *minmax_def  = hdrl_collapse_minmax_parameter_create(1., 1.);
  hdrl_parameter    *mode_def    = hdrl_collapse_mode_parameter_create(10., 1., 0., HDRL_MODE_MEDIAN, 0);
  cpl_parameterlist *pcollapse   = hdrl_collapse_parameter_create_parlist(RECIPE_NAME, "collapse", "MEDIAN", sigclip_def, minmax_def, mode_def) ;
  for (cpl_parameter * p = cpl_parameterlist_get_first(pcollapse); p != NULL; p = cpl_parameterlist_get_next(pcollapse)) {
      cpl_parameterlist_append(self, cpl_parameter_duplicate(p));
  }
  cpl_parameterlist_delete(pcollapse);
  hdrl_parameter_delete(minmax_def);
  hdrl_parameter_delete(sigclip_def);
  hdrl_parameter_delete(mode_def);
  
  par = cpl_parameter_new_value(RECIPE_NAME".bpm.enable", CPL_TYPE_BOOL,
                                " If TRUE the recipe tries to detect and mark bad pixels, e.g. cosmics  \n"
                                " on the single aligned input spectra by following the algorithm        \n"
                                " described in the HDRL manual (Bad-pixel detection on a stack of       \n"
                                " identical images). See also the bpm parameters for the offered methods\n"
                                " and more information. If FALSE the bpm parameters are ignored         \n",
                                RECIPE_NAME, CPL_FALSE);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "bpm.enable");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);


  /* Create BPM_3d parameters */
  hdrl_parameter * deflts = hdrl_bpm_3d_parameter_create(3., 3.,
                                                         HDRL_BPM_3D_THRESHOLD_RELATIVE);
  cpl_parameterlist * bpm_param = hdrl_bpm_3d_parameter_create_parlist(
      RECIPE_NAME, "bpm", deflts) ;
  hdrl_parameter_delete(deflts) ;
  for (cpl_parameter * p = cpl_parameterlist_get_first(bpm_param) ;
      p != NULL; p = cpl_parameterlist_get_next(bpm_param))
    cpl_parameterlist_append(self, cpl_parameter_duplicate(p));
  cpl_parameterlist_delete(bpm_param);


  par = cpl_parameter_new_value(RECIPE_NAME".noIDP", CPL_TYPE_BOOL,
                                " If TRUE, the recipe will not check if the input"
                                " is IDP compliant and will not produce an IDP "
                                "compliant output. Moreover, the user can specify"
                                " the extension and the column names - see below",
                                RECIPE_NAME, CPL_FALSE);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "noIDP");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* --user-specified-column name */
  par = cpl_parameter_new_value(RECIPE_NAME".noIDP.extension",	CPL_TYPE_INT,
                                "Fits extension containing the spectrum table",
                                RECIPE_NAME, 1);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "noIDP.extension");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* --user-specified-column name */
  par = cpl_parameter_new_value(RECIPE_NAME".noIDP.colname-wave",	CPL_TYPE_STRING,
                                " Name of the column containing the wavelength "
                                " for a non IDP input",
                                RECIPE_NAME, "WAVE");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "noIDP.colname-wave");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* --user-specified-column name */
  par = cpl_parameter_new_value(RECIPE_NAME".noIDP.colname-flux",	CPL_TYPE_STRING,
                                " Name of the column containing the flux "
                                " for a non IDP input",
                                RECIPE_NAME, "FLUX");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "noIDP.colname-flux");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* --user-specified-column name */
  par = cpl_parameter_new_value(RECIPE_NAME".noIDP.colname-err",	CPL_TYPE_STRING,
                                " Name of the column containing the flux error"
                                " for a non IDP input",
                                RECIPE_NAME, "ERR");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "noIDP.colname-err");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* --user-specified-column name */
  par = cpl_parameter_new_value(RECIPE_NAME".noIDP.colname-qual",	CPL_TYPE_STRING,
                                " Name of the column containing the quality/bpm"
                                " for a non IDP input",
                                RECIPE_NAME, "QUAL");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "noIDP.colname-qual");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);


  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Function to save the table result in a FITS file
 *
 * @param   spectrum_table    table that need to be save to a fits file
 * @param   parlist           input parameterlist in the recipe
 * @param   frameset          input frameset      in the recipe
 * @param   qclist            input propertylist with quality control parameter
 * @param   raw_frame         input raw frame
 * @param   isIDP             boolean that specify if it's IDP or not
 * @param   filename          Name of the ouput FITS file
 * @param   catg              PRO.CATG in the   FITS file
 * @param   nKeys             Number of enumerate keywords
 * @param   copy_keys         List of keywords for copy in the output file
 * @param   nOutputs          Number of output columns
 * @param   output_columns    List of output_columns
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code save_tab(
    cpl_table                 *spectrum_table,
    const cpl_parameterlist   *parlist,
    cpl_frameset              *frameset,
    cpl_propertylist          *qclist,
    cpl_frame                 *raw_frame,
    cpl_boolean               isIDP,
    const char                *filename,
    const char                *catg,
    cpl_size                  nKeys,
    char                      **copy_keys,
    cpl_size                  nOutputs,
    char                      **output_columns)
{
  cpl_propertylist *header_pri = cpl_propertylist_new();
  cpl_propertylist *header_ext = cpl_propertylist_new();

  cpl_propertylist_append(header_pri, qclist);

  if (!isIDP) {
      /* Add the input PRO.CATG */
      cpl_propertylist_update_string(header_pri, CPL_DFS_PRO_CATG, catg);

      /* Save table using the cpl_dfs table, only for the normal output products */
      cpl_dfs_save_table( frameset,
                          NULL,
                          parlist,
                          frameset,
                          raw_frame,
                          spectrum_table,
                          header_ext,
                          RECIPE_NAME,
                          header_pri,
                          NULL,
                          PACKAGE "/" PACKAGE_VERSION,
                          filename);
  } else {

      cpl_error_code err;

      cpl_frame *product_frame = cpl_frame_new();
      err  = cpl_frame_set_filename(product_frame, filename);
      err |= cpl_frame_set_tag(     product_frame, catg);
      err |= cpl_frame_set_type(    product_frame, CPL_FRAME_TYPE_TABLE);
      err |= cpl_frame_set_group(   product_frame, CPL_FRAME_GROUP_PRODUCT);
      err |= cpl_frame_set_level(   product_frame, CPL_FRAME_LEVEL_FINAL);
      if (err) {
          cpl_frame_delete(product_frame);
          cpl_propertylist_delete(header_pri);
          cpl_propertylist_delete(header_ext);
          return err;
      }

      /* Add the new frame to the frameset to tell esorex that there is a new product */

      /* Add DataFlow keywords. */
      const char *dict_id = "DUMMY_DICT";
      err = cpl_dfs_setup_product_header( header_pri,
                                          product_frame,
                                          frameset,
                                          parlist,
                                          RECIPE_NAME,
                                          PACKAGE "/" PACKAGE_VERSION,
                                          dict_id,
                                          raw_frame);
      if (err) {
          cpl_frame_delete(product_frame);
          cpl_propertylist_delete(header_pri);
          cpl_propertylist_delete(header_ext);
          return err;
      }

      /* Fill IDP headers: Primary/Extension standard keywords */
      cpl_boolean deleteColumnDefinition = CPL_TRUE;
      err = esotk_idp_fill_headers(frameset, spectrum_table, header_pri, header_ext, deleteColumnDefinition, nKeys, copy_keys);
      if (err) {
          cpl_frame_delete(product_frame);
          cpl_propertylist_delete(header_pri);
          cpl_propertylist_delete(header_ext);
          return err;
      }

      /* Add the input PRO.CATG */
      cpl_propertylist_erase(header_pri, CPL_DFS_PRO_CATG);
      cpl_propertylist_update_string(header_pri, CPL_DFS_PRO_CATG, catg);

      /* Create the output table*/
      cpl_table *local_table = append_columns_table(frameset, spectrum_table, nOutputs, output_columns, header_pri, header_ext);
      if (!local_table) {
          cpl_frame_delete(product_frame);
          cpl_propertylist_delete(header_pri);
          cpl_propertylist_delete(header_ext);
          return cpl_error_get_code();
      }

      /* Save table without modify the primary header */
      err = cpl_table_save(local_table, header_pri, header_ext, filename, CPL_IO_CREATE);
      cpl_table_delete(local_table);
      if (err) {
          cpl_frame_delete(product_frame);
          cpl_propertylist_delete(header_pri);
          cpl_propertylist_delete(header_ext);
          return err;
      }

      cpl_frameset_insert(frameset, product_frame);
  }

  /* Cleanup */
  cpl_propertylist_delete(header_pri);
  cpl_propertylist_delete(header_ext);

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Convert array to image
 *
 * @param   arr          array data
 * @param   rej          non-valid positions in the array data
 *
 * @return  cpl_image    Output create with the input data without the reject data
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_image * arr_to_img(
    const cpl_array           *arr,
    const cpl_array           *rej)
{

  const cpl_size size = cpl_array_get_size(arr);
  cpl_image      *img = cpl_image_new(size, 1, HDRL_TYPE_DATA);

  for (cpl_size i = 0; i < size; i++) {

      int    rej_this = 0;
      double el       = cpl_array_get(arr, i, &rej_this);

      /* Reject if either the pixel in arr is rejected or if the corresponding value in rej is different from zero */
      const cpl_boolean qty_reject = (rej != NULL && cpl_array_get_int(rej, i, NULL) != 0);

      if (qty_reject || rej_this) {
          cpl_image_reject(img, i + 1, 1);
      } else {
          cpl_image_set(   img, i + 1, 1, el);
      }
  }

  return img;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Convert input arrays in a spectrum
 *
 * @param   wlens            Array that contains the wavelength data
 * @param   flx_arr          Array that contains the flux data
 * @param   flx_e_arr        Array that contains the flux error data
 * @param   quality          Array that contains the quality data
 *
 * @return  hdrl_spectrum1D  A hdrl_spectrum1D structure that is the combined of the input arrays
 *
 */
/*----------------------------------------------------------------------------*/
static hdrl_spectrum1D * convert_arrs_to_spectrum1D(
    const cpl_array           *wlens,
    const cpl_array           *flx_arr,
    const cpl_array           *flx_e_arr,
    const cpl_array           *quality)
{

  /* Create flux image map */
  cpl_image *flx = arr_to_img(flx_arr, quality);

  /* Create flux error image map */
  cpl_image *flx_e = NULL;
  if (flx_e_arr) {
      flx_e = arr_to_img(flx_e_arr, NULL);
  }

  /* Create spectrum 1D, depending if exist or not flux error image */
  hdrl_spectrum1D *s = NULL;
  if (flx_e) {
      s = hdrl_spectrum1D_create(flx, flx_e, wlens, global_scale);
  } else {
      s = hdrl_spectrum1D_create_error_free(flx, wlens, global_scale);
  }

  cpl_image_delete(flx);
  cpl_image_delete(flx_e);

  return s;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Convert table to spectrum
 *
 * @param   tb               Input table with the column data
 * @param   wln_name         Wavelength column name in the table
 * @param   flx_name         Flux       column name in the table
 * @param   flx_e_name       Flux error column name in the table
 * @param   qual_column      Quality    column name in the table
 *
 * @return  hdrl_spectrum1D  A hdrl_spectrum1D structure that is the combined of the input arrays
 *
 */
/*----------------------------------------------------------------------------*/
static hdrl_spectrum1D * convert_to_spectrum(
    const cpl_table           *tb,
    const char                *wln_name,
    const char                *flx_name,
    const char                *flx_e_name,
    const char                *qual_column)
{

  /* Get Wavelength and flux array */
  const cpl_array *wlens    = cpl_table_get_array(tb, wln_name, 0);
  const cpl_array *flx_cnst = cpl_table_get_array(tb, flx_name, 0);

  const cpl_size  size      = cpl_array_get_size( flx_cnst);
  cpl_array       *flx      = cpl_array_duplicate(flx_cnst);

  /* Using the value of the rej position flux, mark as valid/non-valid value in the array. Also check if the value is NaN */
  for (cpl_size i = 0; i < size; i++) {

      int          rej = 0;
      const double val = cpl_array_get(flx, i, &rej);

      if (!rej) {
          if (!isfinite(val)) {
              cpl_array_set_invalid(flx, i);
          }
      }
  }

  /* Get flux error array */
  const cpl_array *flx_e = NULL;
  if (flx_e_name != NULL) {
      flx_e = cpl_table_get_array(tb, flx_e_name, 0);
  }

  /* Get quality array */
  const cpl_array *quality = NULL;
  if (qual_column != NULL && cpl_table_has_column(tb, qual_column)) {
      quality = cpl_table_get_array(tb, qual_column, 0);
  }

  /* Convert all arrays in a spectrum 1D */
  hdrl_spectrum1D *to_ret = convert_arrs_to_spectrum1D(wlens, flx, flx_e, quality);
  cpl_array_delete(flx);

  return to_ret;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Generate a hdrl_spectrum1Dlist from the input frameset
 *
 * @param   fset                 Frameset
 * @param   wavelength_column    Name of the Wavelength colum
 * @param   flux_column          Name of the Flux c     olum
 * @param   flux_error_column    Name of the Flux error colum
 * @param   qual_column          Name of the Quality    colum
 * @param   rescale              Is necessary to rescale the data?
 * @param   rescale_vector       Returns the value used for scaling
 *
 * @return  hdrl_spectrum1Dlist  Output spectrum 1D list
 *
 */
/*----------------------------------------------------------------------------*/
static hdrl_spectrum1Dlist * generate_list_from_frameset(
    const cpl_frameset        *fset,
    const cpl_boolean         noIDP,
    const int                 extension,
    const char                *wavelength_column,
    const char                *flux_column,
    const char                *flux_error_column,
    const char                *qual_column,
    const cpl_boolean         rescale,
    cpl_vector                *rescale_vector)
{

  const cpl_size      size = cpl_frameset_get_size(fset);
  hdrl_spectrum1Dlist *l   = hdrl_spectrum1Dlist_new();

  /* Generate a spectrum list with the input frameset spectrums */
  for (cpl_size i = 0; i < size; i++) {

      /* Get data values from each frame in the input frameset */
      const cpl_frame *frame = cpl_frameset_get_position_const(fset, i);
      const char      *fname = cpl_frame_get_filename(frame);
      cpl_table       *tb    = cpl_table_load(fname, extension, 0);

      /* Check the mandatory columns in case of IDP input */
      if (noIDP == CPL_FALSE &&
	  (!cpl_table_has_column(tb, wavelength_column)
          || !cpl_table_has_column(tb, flux_column      )
          || !cpl_table_has_column(tb, flux_error_column))) {

          cpl_table_delete(tb);
          hdrl_spectrum1Dlist_delete(l);
          return NULL;
      }

      /* Check the mandatory columns in case of on IDP input */
      if (noIDP == CPL_TRUE &&
	  (!cpl_table_has_column(tb, wavelength_column)
          || !cpl_table_has_column(tb, flux_column      ))) {

          cpl_table_delete(tb);
          hdrl_spectrum1Dlist_delete(l);
          return NULL;
      }

      cpl_size depth = cpl_table_get_column_depth(tb, wavelength_column);

      hdrl_spectrum1D *s = NULL;

      if (noIDP == CPL_FALSE) { /* IDP compliance must be fulfilled in format */
	  /* Convert data in a spectrum 1D */
	  s = convert_to_spectrum(tb, wavelength_column, flux_column, flux_error_column, qual_column);
      } else { /* IDP compliance can be violated and also the error column can be optional */

	  if(cpl_table_has_column(tb, flux_error_column)) { /* error column is found */
	      if(depth > 0) { /* We are in the normal IDP format case */
		  /* Convert data in a spectrum 1D */
		  s = convert_to_spectrum(tb, wavelength_column, flux_column, flux_error_column, qual_column);
	      } else {
		  s = hdrl_spectrum1D_convert_from_table(tb, flux_column, wavelength_column, flux_error_column, qual_column, global_scale);
	      }
	  } else { /* error column is not found, thus supposed to be optional */
	      cpl_msg_info(cpl_func, "No error column found in the table - assuming an error of 0.");
	      if(depth > 0) { /* We are in the normal IDP format case */
		  /* Convert data in a spectrum 1D */
		  s = convert_to_spectrum(tb, wavelength_column, flux_column, NULL, qual_column);
	      } else {
		  s = hdrl_spectrum1D_convert_from_table(tb, flux_column, wavelength_column, NULL, qual_column, global_scale);
	      }
	  }
      }

      /* Add spectrum to the list */
      hdrl_spectrum1Dlist_set(l, s, i);
      cpl_table_delete(tb);
  }

  /* Rescale data */
  if (rescale && size > 0) {

      /* Get median from the first spectrum in the list (reference spectrum) */
      const hdrl_spectrum1D *ref_spectrum1D = hdrl_spectrum1Dlist_get_const(l, 0);
      const hdrl_value      ref_median    = hdrl_image_get_median(hdrl_spectrum1D_get_flux(ref_spectrum1D));

      /* Check that the median value are normal and different of zero */
      if (ref_median.data == 0. || !isfinite(ref_median.data)) {
          hdrl_spectrum1Dlist_delete(l);
          cpl_msg_error(cpl_func,
                        "The first of the spectra has median %g for colum %s. Please disable rescaling or remove the spectrum.",
                        ref_median.data, flux_column);
          return NULL;
      }
      cpl_vector_set(rescale_vector, 0, ref_median.data);
      cpl_msg_info(cpl_func, "Rescale all the spectra according to the median of the first %g +/- %g", ref_median.data, ref_median.error);

      /* Apply the rescale value to each spectrum in the list (except the first) */
      for (cpl_size i = 1; i < size; i++) {

          /* Get spectrum to rescale and it median */
          hdrl_spectrum1D  *spectrum1D = hdrl_spectrum1Dlist_get(l, i);
          const hdrl_value median      = hdrl_image_get_median(hdrl_spectrum1D_get_flux(spectrum1D));

          /* Check that the median value are normal and different of zero */
          if (median.data == 0. || !isfinite(median.data)) {
              hdrl_spectrum1Dlist_delete(l);
              cpl_msg_error(cpl_func,
                            "The %lld-th spectrum has median %g for colum %s. Please disable rescaling or remove the spectrum.",
                            i + 1, median.data, flux_column);
              return NULL;
          }
          cpl_vector_set(rescale_vector, i, median.data);

          /* Apply rescale */
          hdrl_spectrum1D_div_scalar(spectrum1D, median    );
          hdrl_spectrum1D_mul_scalar(spectrum1D, ref_median);

          /* Get new median and info the user */
          const hdrl_value new_median = hdrl_image_get_median(hdrl_spectrum1D_get_flux(spectrum1D));
          cpl_msg_info(cpl_func,
                       "Rescaled the %lld-th spectrum, the median was %g +/- %g, now the median is %g +/- %g.",
                       i + 1, median.data, median.error, new_median.data, new_median.error);
      }
  }

  return l;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Generate wavelenght array form the input frameset
 *
 * @param   fset        Frameset in the recipe
 *
 * @return  cpl_array   Data wavelenght in fset
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_array * generate_wlenghts_array_from_frameset(
    const cpl_frameset        *fset)
{

  /* Get first frame in the frameset */
  const cpl_frame *frame = cpl_frameset_get_position_const(fset, 0);
  const char      *fname = cpl_frame_get_filename(frame);

  /* Get data of the frame */
  cpl_vector      *v     = cpl_vector_load(fname, 0);
  const cpl_size  size   = cpl_vector_get_size(v);
  cpl_array       *arr   = cpl_array_new(size, HDRL_TYPE_DATA);

  /* Generate array with the vector data */
  for (cpl_size i = 0; i < size; i++) {
      double d = cpl_vector_get(v, i);
      cpl_array_set(arr, i, d);
  }

  /* Cleanup */
  cpl_vector_delete(v);

  return arr;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Get wavelenghts from the first destination frame in the frameset
 *
 * @param   fset         Frameset in the recipe
 * @param   list         hdrl spectrum 1D list structure that contains all of input spectrums in the fset
 *
 * @return  cpl_array    Contains all of the wavelengts in the frameset
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_array * get_wlengths(
    const cpl_frameset        *fset,
    const hdrl_spectrum1Dlist *list)
{

  cpl_array *wlengths = NULL;

  /* Get wavelengths */
  if (cpl_frameset_get_size(fset) == 1) {

      /* we have only one frame */
      wlengths = generate_wlenghts_array_from_frameset(fset);
      cpl_msg_info(cpl_func, "SOF contains destination wavelengths.");

  } else {

      /* Get the first spectrum in the list and get the wavelenghts of it */
      const hdrl_spectrum1D *s   = hdrl_spectrum1Dlist_get_const(list, 0);
      const cpl_array       *wls = hdrl_spectrum1D_get_wavelength(s).wavelength;

      /* Create array from the cpl_array "wls" pointer */
      wlengths = cpl_array_duplicate(wls);
      cpl_msg_info(cpl_func, "SOF did not contain destination wavelengths. Wavelengths of the first spectrum will be used instead.");
  }

  return wlengths;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Collapse spectrum 1D list
 *
 * @param   list                          hdrl spectrum 1D list structure that contains all of input spectrums in the fset
 * @param   stacking_par                  hdrl_parameter regulating the stacking
 * @param   wlengths                      wavelengths the resulting spectrum is defined on
 * @param   resample_par                  hdrl_parameter regulating the resampling
 * @param   bpm_par                       hdrl_parameter BPM_3D computation parameters
 * @param   mark_bpm_in_interpolation     if true interpolated pixels whose neighbors (in the original spectrum)
 *                                        are rejected, are not considered during collapsing
 * @param   result                        Output hdrl_spectrum1D result spectrum of the collapse
 * @param   contrib                       Output cpl_image contribution mask result of the collapse
 * @param   resampled_and_aligned_fluxes  Resampled and aligned fluxes to be collapsed
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code hdrl_spectrum1Dlist_collapse_esotk(
    const hdrl_spectrum1Dlist *list,
    const hdrl_parameter      *stacking_par,
    const cpl_array           *wlengths,
    const hdrl_parameter      *resample_par,
    const hdrl_parameter      *bpm_par,
    const cpl_boolean         mark_bpm_in_interpolation,
    hdrl_spectrum1D           **result,
    cpl_image                 **contrib,
    hdrl_imagelist            **resampled_and_aligned_fluxes)
{
  if(bpm_par == NULL) {
    return hdrl_spectrum1Dlist_collapse(list, stacking_par, wlengths,
                                        resample_par, mark_bpm_in_interpolation,
                                        result, contrib, resampled_and_aligned_fluxes);
  }

  hdrl_spectrum1D * irrelevant_res = NULL;
  cpl_image * irrelevant_contrib = NULL;

  hdrl_imagelist * aligned_fluxes = NULL;

  cpl_error_code err = hdrl_spectrum1Dlist_collapse(list, stacking_par, wlengths,
                                                    resample_par, mark_bpm_in_interpolation,
                                                    &irrelevant_res, &irrelevant_contrib, &aligned_fluxes);

  cpl_image_delete(irrelevant_contrib);
  hdrl_spectrum1D_delete(&irrelevant_res);

  if(err){
      hdrl_imagelist_delete(aligned_fluxes);
      return err;
  }

  cpl_imagelist * bpms = hdrl_bpm_3d_compute(aligned_fluxes, bpm_par);

  for(cpl_size i = 0; i < hdrl_imagelist_get_size(aligned_fluxes); ++i){
      hdrl_image * img = hdrl_imagelist_get(aligned_fluxes, i);

      cpl_mask * mask_hdrl = cpl_image_get_bpm(hdrl_image_get_image(img));
      const cpl_image * img_rej = cpl_imagelist_get_const(bpms, i);
      cpl_mask * msk = cpl_mask_threshold_image_create(img_rej, 0.5, 1.5);
      cpl_mask_or(mask_hdrl,  msk);
      cpl_mask_delete(msk);
  }

  cpl_imagelist_delete(bpms);

  hdrl_image * hdrl_img_res = NULL;
  err = hdrl_imagelist_collapse(aligned_fluxes, stacking_par, &hdrl_img_res, contrib);

  hdrl_imagelist_delete(aligned_fluxes);

  if(err){
      hdrl_image_delete(hdrl_img_res);
      return err;
  }

  const hdrl_spectrum1D_wave_scale scale =
      hdrl_spectrum1D_get_scale(hdrl_spectrum1Dlist_get_const(list, 0));

  *result = hdrl_spectrum1D_create(hdrl_image_get_image(hdrl_img_res),
                                   hdrl_image_get_error(hdrl_img_res),
                                   wlengths,
                                   scale);

  hdrl_image_delete(hdrl_img_res);

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Function that collapse all of spectra in the input list
 *
 * @param   list            hdrl spectrum 1D list structure that contains all of input spectrums in the fset
 * @param   wlengths        All of the wavelengths
 * @param   remove_rej      Remove non-valid values?
 * @param   collapse_par    hdrl_parameter with the values needed in the collapse
 * @param   interp_par      hdrl_parameter with the values needed in the interpolation
 * @param   arg_spectrum1D  Output hdrl_spectrum1D   result of the collapse
 * @param   arg_contrib     Output cpl_image contrib result of the collapse
 *
 */
/*----------------------------------------------------------------------------*/
static void collapse_spectra(
    const hdrl_spectrum1Dlist *list,
    const cpl_array           *wlengths,
    const cpl_boolean         remove_rej,
    const hdrl_parameter      *collapse_par,
    const hdrl_parameter      *interp_par,
    const hdrl_parameter      *bpm_par,
    hdrl_spectrum1D           **arg_spectrum1D,
    cpl_image                 **arg_contrib)
{

  /* local variables */
  hdrl_imagelist  *aligned_fluxes = NULL;
  hdrl_spectrum1D *spectrum1D     = NULL;
  cpl_image       *contrib        = NULL;

  /* Collapse spectrum 1D list */
  cpl_error_code err = hdrl_spectrum1Dlist_collapse_esotk(
      list, collapse_par, wlengths, interp_par, bpm_par, remove_rej, &spectrum1D, &contrib, &aligned_fluxes);

  /* Cleanup the unnecessary information aligned_fluxes cpl_image */
  hdrl_imagelist_delete(aligned_fluxes);

  /* Check output result in the collapse */
  if (err || !spectrum1D || !contrib) {
      hdrl_spectrum1D_delete(&spectrum1D);
      cpl_image_delete(contrib);
      contrib    = NULL;
      spectrum1D = NULL;
      cpl_error_set_message(cpl_func, err, "Combining returned an error");
  }

  /* return the output values */
  *arg_contrib    = contrib;
  *arg_spectrum1D = spectrum1D;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Function that collapse all of spectra in the input list and convert in a table
 *
 * @param   list             hdrl spectrum 1D list structure that contains all of input spectrums in the fset
 * @param   wlengths         All of the wavelengths
 * @param   remove_rej       Remove non-valid values?
 * @param   collapse_par     hdrl_parameter with the values needed in the collapse
 * @param   interp_par       hdrl_parameter with the values needed in the interpolation
 * @param   wav_cl_name      Wavelength    column name in the table
 * @param   flx_cl_name      Flux          column name in the table
 * @param   flx_e_cl_name    Flux error    column name in the table
 * @param   bpm_cl_name      Bad pixel map column name in the table
 * @param   contrib_cl_name  contrib       column name in the table
 * @param   snr_cl_name      Signal/Noise  column name in the table
 *
 * @return  cpl_table        Output table that contains all the data
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_table * collapse_and_convert_to_table(
    const hdrl_spectrum1Dlist *list,
    const cpl_array           *wlengths,
    const cpl_boolean         remove_rej,
    const hdrl_parameter      *collapse_par,
    const hdrl_parameter      *interp_par,
    const hdrl_parameter      *bpm_par,
    const char                *wav_cl_name,
    const char                *flx_cl_name,
    const char                *flx_e_cl_name,
    const char                *bpm_cl_name,
    const char                *contrib_cl_name,
    const char                *snr_cl_name)
{
  cpl_table *tb = NULL;

  /* If list exist fill tb */
  if (list) {

      /* Collapse spectrum1Dlist in only 1 spectrum1D */
      hdrl_spectrum1D *spectrum1D = NULL;
      cpl_image       *contrib    = NULL;
      collapse_spectra(list, wlengths, remove_rej, collapse_par, interp_par, bpm_par, &spectrum1D, &contrib);

      /* If the collapse is successfully */
      if (spectrum1D && contrib) {

          /* Convert the spectrum in a table */
          tb = hdrl_spectrum1D_convert_to_table(spectrum1D, flx_cl_name, wav_cl_name, flx_e_cl_name, bpm_cl_name);

          if (!strcmp(bpm_cl_name, "BPM_REDUCED")) {
              cpl_table_set_column_unit(tb, bpm_cl_name, ESOTK_SPECTRUM1D_BPM_REDUCED_TUNIT_VALUE);
          } else {
              cpl_table_set_column_unit(tb, bpm_cl_name, ESOTK_SPECTRUM1D_BPM_TUNIT_VALUE);
          }

          /* Put the contrib image in the column contrib of the table specified in the input parameter */
          cpl_table_new_column(tb, contrib_cl_name, CPL_TYPE_INT);
          cpl_table_set_column_unit(tb, contrib_cl_name, ESOTK_SPECTRUM1D_CONTRIB_TUNIT_VALUE);
          for (cpl_size i = 0; i < cpl_image_get_size_x(contrib); i++) {

              /* Get the contrib i-th position data */
              int rej = 0;
              const int contribution = (int)cpl_image_get(contrib, i + 1, 1, &rej);

              /* Set the position in the table */
              cpl_table_set_int(tb, contrib_cl_name, i, contribution);
          }

          /* Put the signal/noise data in the column name signal/noise of the table specified in the input parameter */
          cpl_table_new_column(tb, snr_cl_name, CPL_TYPE_DOUBLE);
          cpl_table_set_column_unit(tb, snr_cl_name, ESOTK_SPECTRUM1D_SNR_TUNIT_VALUE);
          for (cpl_size i = 0; i < hdrl_spectrum1D_get_size(spectrum1D); i++) {

              /* Get the flux in the i-th position data */
              int rej;
              const hdrl_value val = hdrl_spectrum1D_get_flux_value(spectrum1D, i, &rej);

              /* If the data are valid, calculate the signal/noise */
              if (!rej) {
                  const double d = val.data / val.error;
                  if(isfinite(d))
                    cpl_table_set_double(tb, snr_cl_name, i, d);
              }
          }

      }

      /* Cleanup */
      hdrl_spectrum1D_delete(&spectrum1D);
      cpl_image_delete(contrib);
  }

  return tb;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Add columns to the input table
 *
 * @param   t           Input/Output table that at the end contains all columns in "other"
 * @param   other       Input table for copy the columns
 *
 */
/*----------------------------------------------------------------------------*/
static void append_table(
    cpl_table                 *t,
    const cpl_table           *other)
{

  /* Get column names in the other table */
  cpl_array *col_names = cpl_table_get_column_names(other);

  /* Insert the columns that not appear in the "t" table */
  for (cpl_size i = 0; i < cpl_array_get_size(col_names); i++) {

      const char *name = cpl_array_get_string(col_names, i);
      if(!cpl_table_has_column(t, name)) {
          cpl_table_duplicate_column(t, name, other, name);
      }
  }

  cpl_array_delete(col_names);
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Generate an output cpl_table and update the header properly with the frameset and the input table
 *
 * @param   frameset      input raw frameset
 * @param   source        cpl_table with the data
 * @param   size          size of the array of column names
 * @param   out_cols      array of string with the names of the columns in the output
 * @param   pri_header    in/out header of extension
 * @param   ext_header    in/out header of extension
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_table * append_columns_table(
    cpl_frameset              *frameset,
    const cpl_table           *source,
    cpl_size                  size,
    char                      **out_cols,
    cpl_propertylist          *pri_header,
    cpl_propertylist          *ext_header)
{
  /* Check null inputs */
  cpl_error_ensure(frameset && source && out_cols && ext_header, CPL_ERROR_NULL_INPUT,
                   return NULL, "Null inputs");

  /* Get the extension header in the first raw frame of the frameset */
  cpl_propertylist *header_raw = NULL;
  const char *filename = NULL;
  cpl_size nColumns_raw = 0;
  cpl_size  nframes = cpl_frameset_get_size(frameset);
  for (cpl_size i = 0; i < nframes && !header_raw; i++) {
      const cpl_frame *cur_frame = cpl_frameset_get_position(frameset, i);
      if (cpl_frame_get_group(cur_frame) == CPL_FRAME_GROUP_RAW) {
          filename   = cpl_frame_get_filename(cur_frame);
          header_raw = cpl_propertylist_load(filename, 1);
          cpl_table *data = cpl_table_load(filename, 1, 0);
          nColumns_raw = cpl_table_get_ncol(data);
          cpl_table_delete(data);
      }
  }

  if(header_raw == NULL || filename == NULL || nColumns_raw < 1){
      cpl_propertylist_delete(header_raw);
      cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                            "There isn't a extension header for a raw cpl_frame in the cpl_frameset");
      return NULL;
  }




  /* Check if all_token */
  cpl_boolean all_token = CPL_FALSE;
  if (size == 1 && !strcmp(out_cols[0], "ALL")) {
      /* Set all variable */
      all_token = CPL_TRUE;
  } else {

      /* Check that all columns in the out_cols are inside the input table */
      for (cpl_size i = 0; i < size; i++) {

          char *name = NULL;
          if (       !strcmp(out_cols[i], "QUAL"        )){
              name = cpl_sprintf("%s", "BPM");
          } else if (!strcmp(out_cols[i], "QUAL_REDUCED")) {
              name = cpl_sprintf("%s", "BPM_REDUCED");
          } else {
              name = cpl_sprintf("%s", out_cols[i]);
          }

          if (!cpl_table_has_column(source, name)) {
              cpl_free(name);
              cpl_propertylist_delete(header_raw);
              cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND, "It cannot exist column=%s in the output table", out_cols[i]);
              return NULL;
          }
          cpl_free(name);
      }
  }

  /* Variable to return: output table and index of columns */
  cpl_table *target = cpl_table_new(cpl_table_get_nrow(source));
  cpl_size  idx     = 1;

  /* Loop over the extension header of the raw cpl_frame */
  for (cpl_size col = 0; col < nColumns_raw; col++) {

      cpl_boolean isWave_col = (col == 0) ? CPL_TRUE : CPL_FALSE;

      /* Create target names */
      char *ttype_in  = cpl_sprintf("TTYPE%lld",  col + 1);
      char *tunit_in  = cpl_sprintf("TUNIT%lld",  col + 1);
      char *tutyp_in  = cpl_sprintf("TUTYP%lld",  col + 1);
      char *tucd_in   = cpl_sprintf("TUCD%lld",   col + 1);

      /* Get name column */
      const char *name = cpl_propertylist_get_string(header_raw, ttype_in);

      /* Check if exist in the list */
      cpl_boolean exist_name = CPL_FALSE;
      if (all_token || isWave_col) {
          exist_name = CPL_TRUE;
      } else {
          for (cpl_size i = 0; i < size && !exist_name; i++) {
              if (!strcmp(out_cols[i], name)) {
                  exist_name = CPL_TRUE;
              }
          }
      }

      /* Add to the output table */
      if (exist_name) {

          cpl_boolean exist_column = cpl_table_has_column(source, name);
          if (exist_column) {

              /* Copy column */
              cpl_table_duplicate_column(target, name, source, name);

              /* Create target names */
              char *tunit_out = cpl_sprintf("TUNIT%lld",  idx);
              char *tutyp_out = cpl_sprintf("TUTYP%lld",  idx);
              char *tucd_out  = cpl_sprintf("TUCD%lld",   idx);

              /* Get values from the extension header */
              char *tunit = NULL;
              if(cpl_propertylist_has(header_raw, tunit_in)) {
        	  tunit = (char *)cpl_propertylist_get_string(header_raw, tunit_in);
              } else {
        	  tunit = cpl_sprintf("%s",  "");
              }
              const char *tutyp = cpl_propertylist_get_string(header_raw, tutyp_in);
              const char *tucd  = cpl_propertylist_get_string(header_raw, tucd_in );

              /* Update the unit in the output column of the target directly by the cpl_frame raw */
              cpl_table_set_column_unit(target, name, tunit);

              /* Update values */
              cpl_propertylist_update_string(ext_header, tunit_out, tunit);
              cpl_propertylist_update_string(ext_header, tutyp_out, tutyp);
              cpl_propertylist_update_string(ext_header, tucd_out,  tucd );

              /* Get comments from the extension header */
              char *tunit_comment = NULL;
              if(cpl_propertylist_has(header_raw, tunit_in)) {
        	  tunit_comment = (char *)cpl_propertylist_get_comment(header_raw, tunit_in);
              } else {
        	  tunit_comment = cpl_sprintf("%s",  "");
              }

              const char *tutyp_comment = cpl_propertylist_get_comment(header_raw, tutyp_in);
              const char *tucd_comment  = cpl_propertylist_get_comment(header_raw, tucd_in );

              /* Update comments */
              cpl_propertylist_set_comment(ext_header, tunit_out, tunit_comment);
              cpl_propertylist_set_comment(ext_header, tutyp_out, tutyp_comment);
              cpl_propertylist_set_comment(ext_header, tucd_out, tucd_comment);

              /* Cleanup */
              if(!cpl_propertylist_has(header_raw, tunit_in)) {
        	  cpl_free(tunit);
        	  cpl_free(tunit_comment);
              }
              cpl_free(tunit_out);
              cpl_free(tutyp_out);
              cpl_free(tucd_out);

              idx++;
          }
      }

      /* Cleanup */
      cpl_free(ttype_in);
      cpl_free(tunit_in);
      cpl_free(tutyp_in);
      cpl_free(tucd_in);
  }
  cpl_propertylist_delete(header_raw);

  /* Insert columns, not put in the frameset loop */
  cpl_array *col_names = cpl_table_get_column_names(source);
  cpl_size size_cols = cpl_array_get_size(col_names);
  for (cpl_size col = 0; col < size_cols; col++) {

      /* Get column name in the table */
      const char *name = cpl_array_get_string(col_names, col);

      /* Check if this column is previously added in the frameset loop */
      cpl_boolean exist_column = cpl_table_has_column(target, name);
      if (!exist_column) {

          /* Check if this not added column, exist in the list */
          cpl_boolean exist_name = CPL_FALSE;
          if (all_token) {
              exist_name = CPL_TRUE;
          } else {
              for (cpl_size i = 0; i < size && !exist_name; i++) {
                  cpl_msg_info(cpl_func, "out_cols[%lld]=%s, nameCol[%lld]=%s", i, out_cols[i], col, name);
                  if (    !strcmp(out_cols[i],  name         )
                      || (!strcmp(out_cols[i], "QUAL"        ) && !strcmp("BPM",         name))
                      || (!strcmp(out_cols[i], "QUAL_REDUCED") && !strcmp("BPM_REDUCED", name))) {
                      exist_name = CPL_TRUE;
                  }
              }
          }

          /* Add to the output table */
          if (exist_name) {

              /* Copy column */
              cpl_table_duplicate_column(target, name, source, name);

              /* Create target names */
              char *ttype_out = cpl_sprintf("TTYPE%lld",  idx);
              char *tunit_out = cpl_sprintf("TUNIT%lld",  idx);
              char *tutyp_out = cpl_sprintf("TUTYP%lld",  idx);
              char *tucd_out  = cpl_sprintf("TUCD%lld",   idx);

              /* Update values */

              /* Set Unit value */
              const char *unit_str = cpl_table_get_column_unit(source, name);
              if (unit_str){
                  cpl_propertylist_update_string(ext_header, tunit_out, unit_str);
              } else {
                  /* If return NULL, you need to update also the unit in the column table, to be sure that cpl_dfs_save_table() write the value */
                  cpl_propertylist_update_string(ext_header, tunit_out, "");
                  cpl_table_set_column_unit(target, name, "");
              }

              const char *voclass = NULL;
              if (cpl_propertylist_has(ext_header, ESOTK_IDP_EXT_VOCLASS)) {
                  voclass = cpl_propertylist_get_string(ext_header, ESOTK_IDP_EXT_VOCLASS);
              } else {
                  cpl_msg_warning(cpl_func, "The header keyword %s could not "
                                  "be found in the extension.", ESOTK_IDP_EXT_VOCLASS);
                  voclass = cpl_sprintf(" "); /* To avoid a segfault later on */
                  /* This fallback solution has been discussed with the IDP group*/
              }

              /* Set TUTYP-ith and TUCD-ith */
              if (!strcmp(name, "BPM")) {

                  /* Also change the name of the column from "BPM" to "QUAL" */
                  cpl_table_name_column(target, "BPM", ESOTK_SPECTRUM1D_BPM_TTYPE_VALUE);
                  if (!strcmp(voclass, "SPECTRUM v1.0")) {
                      cpl_propertylist_update_string(ext_header, tutyp_out, ESOTK_SPECTRUM1D_BPM_TUTYP_VALUE_VOCLASS_v1);
                  } else {
                      cpl_propertylist_update_string(ext_header, tutyp_out, ESOTK_SPECTRUM1D_BPM_TUTYP_VALUE_VOCLASS_v2);
                  }
                  cpl_propertylist_update_string(ext_header, tucd_out,  ESOTK_SPECTRUM1D_BPM_TUCD_VALUE );

              } else if (!strcmp(name, "BPM_REDUCED")) {

                  /* Also change the name of the column from "BPM_REDUCED" to "QUAL_REDUCED" */
                  cpl_table_name_column(target, "BPM_REDUCED", ESOTK_SPECTRUM1D_BPM_REDUCED_TTYPE_VALUE);
                  if (!strcmp(voclass, "SPECTRUM v1.0")) {
                      cpl_propertylist_update_string(ext_header, tutyp_out, ESOTK_SPECTRUM1D_BPM_REDUCED_TUTYP_VALUE_VOCLASS_v1);
                  } else {
                      cpl_propertylist_update_string(ext_header, tutyp_out, ESOTK_SPECTRUM1D_BPM_REDUCED_TUTYP_VALUE_VOCLASS_v2);
                  }
                  cpl_propertylist_update_string(ext_header, tucd_out,  ESOTK_SPECTRUM1D_BPM_REDUCED_TUCD_VALUE );

              } else if (!strcmp(name, "CONTRIB") || !strcmp(name, "CONTRIB_REDUCED")) {

                  cpl_propertylist_update_string(ext_header, tutyp_out, ESOTK_SPECTRUM1D_CONTRIB_TUTYP_VALUE);
                  cpl_propertylist_update_string(ext_header, tucd_out,  ESOTK_SPECTRUM1D_CONTRIB_TUCD_VALUE);

              } else if (!strcmp(name, "SNR")) {

                  cpl_propertylist_update_string(ext_header, tutyp_out, ESOTK_SPECTRUM1D_SNR_TUTYP_VALUE);
                  cpl_propertylist_update_string(ext_header, tucd_out,  ESOTK_SPECTRUM1D_SNR_TUCD_VALUE);

              } else if (!strcmp(name, "SNR_REDUCED")) {

                  cpl_propertylist_update_string(ext_header, tutyp_out, ESOTK_SPECTRUM1D_SNR_REDUCED_TUTYP_VALUE);
                  cpl_propertylist_update_string(ext_header, tucd_out,  ESOTK_SPECTRUM1D_SNR_REDUCED_TUCD_VALUE);

              } else {

                  cpl_propertylist_update_string(ext_header, tutyp_out, "");
                  cpl_propertylist_update_string(ext_header, tucd_out,  "");
              }

              /* Set comments */
              /* cpl_propertylist_set_comment(ext_header, ttype_out, ESOTK_IDP_EXT_TTYPE_COMMENT); */
              cpl_propertylist_set_comment(ext_header, tunit_out, ESOTK_IDP_EXT_TUNIT_COMMENT);
              cpl_propertylist_set_comment(ext_header, tutyp_out, ESOTK_IDP_EXT_TUTYP_COMMENT);
              cpl_propertylist_set_comment(ext_header, tucd_out,  ESOTK_IDP_EXT_TUCD_COMMENT );

              /* Cleanup */
              cpl_free(ttype_out);
              cpl_free(tunit_out);
              cpl_free(tutyp_out);
              cpl_free(tucd_out);

              idx++;
          }
      }
  }

  cpl_array_delete(col_names);

  /* Update the keywords related with the 'WAVE' units in the headers */
  if (esotk_idp_update_wave_unit(pri_header, ext_header) != CPL_ERROR_NONE) {
      return NULL;
  }

  return target;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Combine two input tables
 *
 * @param   t1           table 1
 * @param   t2           table 2
 *
 * @return  cpl_table    the output combined table
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_table * attach_tables(
    const cpl_table           *t1,
    const cpl_table           *t2)
{
  /* Check if some of the tables are empty */
  if (t1 == NULL) return cpl_table_duplicate(t2);
  if (t2 == NULL) return cpl_table_duplicate(t1);

  /* Create a new table */
  cpl_size  nrows   = cpl_table_get_nrow(t1);
  cpl_table *to_ret = cpl_table_new(nrows);

  /* Combined t1 and t2 in the output table */
  append_table(to_ret, t1);
  append_table(to_ret, t2);

  return to_ret;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Convert a column in the input table to array
 *
 * @param   tb         Input table
 * @param   cl_name    Column name
 *
 * @return  cpl_array  Data values in the column "cl_name" in the input table "tb"
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_array * convert_to_array(
    const cpl_table           *tb,
    const char                *cl_name)
{

  /* Create a new array with size equal to the input table column */
  const cpl_size size = cpl_table_get_nrow(tb);
  cpl_array      *arr = cpl_array_new(size, cpl_table_get_column_type(tb, cl_name));

  /* Fill the array with the valid values (non-rejected) */
  for (cpl_size i = 0; i < size; i++) {

      int    rej = 0;
      double v   = cpl_table_get(tb, cl_name, i, &rej);

      if (!rej) {
          cpl_array_set(arr, i, v);
      }
  }

  return arr;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Change table NxM by 2xN, where the first row are names and the second row are data arrays
 *
 * @param   tb          Input table NxM
 *
 * @return  cpl_table   Conversion to array table of the input table (2xN)
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_table * convert_to_idp(
    const cpl_table           *tb)
{
  /* Create a new IDP table */
  cpl_array *col_names = cpl_table_get_column_names(tb);
  cpl_table *to_ret    = cpl_table_new(1);

  /* For each column convert a column in array --> Change the table NxM by 2xM
   * with the first row names and the second an array with values */
  for (cpl_size i = 0; i < cpl_array_get_size(col_names); i++) {

      /* Get name column and convert data in array */
      const char *name = cpl_array_get_string(col_names, i);
      cpl_array  *col  = convert_to_array(tb, name);

      /* Add array to the return table */
      cpl_table_new_column_array(to_ret, name, cpl_array_get_type(col), cpl_array_get_size(col));

      /* Set unit column */
      if (       !strcmp(name, "BPM")) {

          cpl_table_set_column_unit(to_ret, name, ESOTK_SPECTRUM1D_BPM_TUNIT_VALUE);

      } else if (!strcmp(name, "BPM_REDUCED")) {

          cpl_table_set_column_unit(to_ret, name, ESOTK_SPECTRUM1D_BPM_REDUCED_TUNIT_VALUE);

      } else if (!strcmp(name, "CONTRIB") || !strcmp(name, "CONTRIB_REDUCED")) {

          cpl_table_set_column_unit(to_ret, name, ESOTK_SPECTRUM1D_CONTRIB_TUNIT_VALUE);

      } else if (!strcmp(name, "SNR")) {

          cpl_table_set_column_unit(to_ret, name, ESOTK_SPECTRUM1D_SNR_TUNIT_VALUE);

      } else if (!strcmp(name, "SNR_REDUCED")) {

          cpl_table_set_column_unit(to_ret, name, ESOTK_SPECTRUM1D_SNR_REDUCED_TUNIT_VALUE);

      } else {

          /* Set as empty string */
          cpl_table_set_column_unit(to_ret, name, "");
      }

      cpl_table_set_array(to_ret, name, 0, col);
      cpl_array_delete(col);
  }

  /* Cleanup */
  cpl_array_delete(col_names);

  return to_ret;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Split the input str string using the delimiter delim. Also get the number of tokens.
 *
 * @param   isKeyword   String management like a keyword
 * @param   str         Input string
 * @param   delim       Delimiter, usually ','
 * @param   numtokens   Number of output tokens
 *
 * @return  array of string with the tokenize str input.
 *
 */
/*----------------------------------------------------------------------------*/
static char ** strsplit(
    cpl_boolean               isKeyword,
    const char                *str,
    const char                *delim,
    cpl_size                  *numtokens)
{
  char     *s           = cpl_strdup(str);
  char     *rest        = s;

  cpl_size tokens_alloc = 1;
  cpl_size tokens_used  = 0;
  char     **tokens     = cpl_calloc(tokens_alloc, sizeof(char *));

  char *token;
  while ((token = strsep(&rest, delim)) != NULL) {

      if (tokens_used == tokens_alloc) {
          tokens_alloc *= 2;
          tokens = cpl_realloc(tokens, tokens_alloc * sizeof(char *));
      }

      if (isKeyword) {
          tokens[tokens_used++] = cpl_strdup(str_trim_and_replace(token, '.', ' '));
      } else {
          tokens[tokens_used++] = cpl_strdup(token);
      }
  }

  if (tokens_used == 0) {
      cpl_free(tokens);
      tokens = NULL;
  } else {
      tokens = cpl_realloc(tokens, tokens_used * sizeof(char *));
  }

  *numtokens = tokens_used;
  cpl_free(s);

  return tokens;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Trim the input str with spaces and replace all characters c_orig by c_replace
 *
 * @param   isKeyword   String management like a keyword
 * @param   str         Input/output string
 * @param   c_orig      character that need replacing, in the keywords usually '.'
 * @param   c_replace   new character in the string, in the keywords usually ' '
 *
 * @return  pointer to the str modified
 *
 */
/*----------------------------------------------------------------------------*/
static char * str_trim_and_replace(
    char                      *str,
    const char                c_orig,
    const char                c_replace)
{
  char *end;

  // Trim leading space
  while(isspace((unsigned char)*str)) str++;

  if (*str == 0) {  // All spaces?
      return str;
  }

  // Trim trailing space
  end = str + strlen(str) - 1;
  while(end > str && isspace((unsigned char)*end)) end--;

  // Write new null terminator character
  end[1] = '\0';

  while(end > str) {
      if ((unsigned char)*end == c_orig) {
          *end = c_replace;
      }
      end--;
  }

  return str;
}

/** @endcond */


/**@}*/
